@@ -49,10 +49,16 @@ extern aCache *cachetop;
 #ifdef DEBUGMODE
 extern void report_fds(aClient *);
 #endif
+extern char *oflagtotext(int oflags); /* For stats o */
+extern char *cflagtotext(int cflags, int uflags); /* For stats c */
+extern char *iflagtotext(int iflags); /* For stats i */
+extern char *pflagtotext(int pflags); /* For stats P */
+extern int report_spamfilters(aClient *cptr, aClient *sptr, int parc, char *parv[]); /* For stats S */
 
 /* internal function defines */
 
 static void show_opers(aClient *, char *);
+static void show_ports(aClient *, char *);
 static void show_servers(aClient *, char *);
 #ifdef DEBUGMODE
 static void send_usage(aClient *, char *);
@@ -61,10 +67,6 @@ static void serv_info(aClient *, char *);
 static void tstats(aClient *, char *);
 
 /* support functions */
-/* show_opers
- * replies to stats p requests
- */
-
 
 #ifdef DEBUGMODE
 static void
@@ -162,7 +164,9 @@ send_usage(aClient *cptr, char *nick)
 }
 #endif  /* DEBUGMODE */
 
-
+/* show_opers
+ * replies to stats P requests
+ */
 static void 
 show_opers(aClient *cptr, char *name) 
 {
@@ -180,7 +184,11 @@ show_opers(aClient *cptr, char *name)
         {
         sendto_one(cptr, ":%s %d %s :%s (%s@%s) Idle: %ld",
                me.name, RPL_STATSDEBUG, name, cptr2->name,
-               cptr2->user->username, cptr2->user->host,
+               cptr2->user->username,
+#ifdef USER_HOSTMASKING
+               IsUmodeH(cptr2)?cptr2->user->mhost:
+#endif
+               cptr2->user->host,
                (long)(timeofday - cptr2->user->last));
         j++;
         }
@@ -198,6 +206,35 @@ show_opers(aClient *cptr, char *name)
            name, j, (j == 1) ? "" : "s");
 }
 
+/* show_ports
+ * replies to stats p requets
+ */
+static void
+show_ports(aClient *cptr, char *name)
+{
+    aPort *tmp;
+    int j = 0;
+
+    if(!ports)
+        sendto_realops("Lost all port configurations!");
+    for(tmp = ports; tmp; tmp = tmp->next)
+    {
+        if(tmp->lstn)
+        {
+            if(IsULine(cptr) || (MyClient(cptr) && IsAdmin(cptr)))
+                sendto_one(cptr,":%s %d %s :%s %s %i %s %i", me.name, RPL_STATSDEBUG,
+                       name, tmp->address?tmp->address:"*", tmp->allow?tmp->allow:"*",
+                       tmp->port, pflagtotext(tmp->flags), tmp->lstn->clients);
+            else
+                sendto_one(cptr,":%s %d %s :<masked> <masked> %i %s %i", me.name, RPL_STATSDEBUG,
+                       name, tmp->port, pflagtotext(tmp->flags), tmp->lstn->clients);
+            j++;
+        }
+    }
+    sendto_one(cptr, ":%s %d %s :%d PORT%s", me.name, RPL_STATSDEBUG,
+           name, j, (j == 1) ? "" : "s");
+}
+
 /* show_servers
  * replies to stats v requests
  */
@@ -536,7 +573,7 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
             sincetime = (acptr->since > timeofday) ? 0 : 
                                 timeofday - acptr->since;
             sendto_one(sptr, Lformat, me.name, RPL_STATSLINKINFO, parv[0],
-                        get_client_name(acptr, TRUE),
+                        get_client_name(acptr, (IsUmodeH(acptr) && !IsAnOper(sptr))?HIDEME:TRUE),
                         (int) SBufLength(&acptr->sendQ),
                         (int) acptr->sendM, (int) acptr->sendK,
                         (int) acptr->receiveM, (int) acptr->receiveK,
@@ -546,7 +583,11 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
         }
         else
         {
-            for (i = 0; i <= highest_fd; i++)
+            if(!(confopts & FLAGS_SHOWLINKS) && !IsAnOper(sptr))
+            {
+                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
+            }
+            else for (i = 0; i <= highest_fd; i++)
             {
                 if (!(acptr = local[i]))
                     continue;
@@ -596,8 +637,8 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
                            sptr->name, "C", tmp->host, tmp->name, tmp->port,
                            tmp->class->name);
                     sendto_one(sptr, rpl_str(RPL_STATSNLINE), me.name,
-                           sptr->name, "N", tmp->host, tmp->name, tmp->flags,
-                           tmp->uflags, tmp->class->name);
+                           sptr->name, "N", tmp->host, tmp->name, cflagtotext(tmp->flags, tmp->uflags),
+                           tmp->class->name);
                 }
                 else
                 {
@@ -605,8 +646,8 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
                                sptr->name, "C", "*", tmp->name, tmp->port,
                                tmp->class->name);
                     sendto_one(sptr, rpl_str(RPL_STATSNLINE), me.name,
-                               sptr->name, "N", "*", tmp->name, tmp->flags,
-                               tmp->uflags, tmp->class->name);
+                               sptr->name, "N", "*", tmp->name, cflagtotext(tmp->flags, tmp->uflags),
+                               tmp->class->name);
                 }
             }
         }
@@ -722,7 +763,7 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
                     continue;
                 sendto_one(sptr, rpl_str(RPL_STATSILINE), me.name,
                            sptr->name, (tmp->legal == -1 ? "Ix" : "I"),
-                           tmp->ipmask, tmp->flags, tmp->hostmask, tmp->port,
+                           tmp->ipmask, iflagtotext(tmp->flags), tmp->hostmask, tmp->port,
                            tmp->class->name);
             }
             break;
@@ -766,13 +807,13 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
         case 'N':
         case 'n':
             sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
-                        "User Connects Today: ", Count.today);
+                        "User Connects Today:", Count.today);
             sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
-                        "User Connects past week: ", Count.weekly);
+                        "User Connects past week:", Count.weekly);
             sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
-                        "User Connects past month: ", Count.monthly);
+                        "User Connects past month:", Count.monthly);
             sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0],
-                        "User Connects past year: ", Count.yearly);
+                        "User Connects past year:", Count.yearly);
             break;
         case 'o':
         case 'O':
@@ -787,7 +828,7 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
                     for(i = 0; tmp->hosts[i]; i++)
                         sendto_one(sptr, rpl_str(RPL_STATSOLINE), me.name,
                                 sptr->name, (tmp->legal == -1 ? "Ox" : "O"),
-                                tmp->hosts[i], tmp->nick, tmp->flags,
+                                tmp->hosts[i], tmp->nick, oflagtotext(tmp->flags),
                                 tmp->class->name);
             }
             else
@@ -797,7 +838,7 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
                     if (tmp->legal == -1)
                         continue;
                     sendto_one(sptr, rpl_str(RPL_STATSOLINE), me.name,
-                            sptr->name, "O", "*", tmp->nick, tmp->flags,
+                            sptr->name, "O", "*", tmp->nick, oflagtotext(tmp->flags),
                             tmp->class->name);
                 }
             }
@@ -805,9 +846,14 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
         }
 
         case 'p':
-        case 'P':
             show_opers(sptr, parv[0]);
             break;
+        case 'P':
+            if(IsAnOper(sptr) || IsULine(sptr))
+                show_ports(sptr, parv[0]);
+            else
+                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
+            break;
 
         case 'Q':
             if(IsAnOper(sptr))
@@ -836,6 +882,12 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
             break;
 
         case 'S':
+            if (IsAnOper(sptr))
+                report_spamfilters(cptr, sptr, parc, parv);
+            else
+                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
+            break;
+
         case 's':
             if (IsAnOper(sptr))
                 list_scache(cptr, sptr, parc, parv);
@@ -883,7 +935,11 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
 
         case 'v':
         case 'V':
-            show_servers(sptr, parv[0]);
+            if(!(confopts & FLAGS_SHOWLINKS) && !IsAnOper(sptr))
+            {
+                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
+            }
+            else show_servers(sptr, parv[0]);
             break;
 
 #ifdef DEBUGMODE
@@ -925,7 +981,11 @@ int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
             break;
 
         case '?':
-            serv_info(sptr, parv[0]);
+            if(!(confopts & FLAGS_SHOWLINKS) && !IsAnOper(sptr))
+            {
+                sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
+            }
+            else serv_info(sptr, parv[0]);
             break;
 
         default:
