@@ -437,7 +437,7 @@ void sendto_alias(AliasInfo *ai, aClient *from, char *pattern, ...)
 
     /* use shortforms only for non-super servers or capable super servers */
     if (!IsULine(to) || ((confopts & FLAGS_SERVHUB)
-                         && (to->serv->uflags & ULF_SFDIRECT)))
+                         && to->serv && (to->serv->uflags & ULF_SFDIRECT)))
         len = ircsprintf(sendbuf, ":%s %s :", from->name, ai->shortform);
     else
 #ifdef PASS_SERVICES_MSGS
@@ -507,10 +507,16 @@ static inline int prefix_buffer(int remote, aClient *from, char *prefix,
                 for(p = user->username; *p; p++)
                     buffer[sidx++] = *p;
             }
-            if (*user->host && !MyConnect(from)) 
+            if (*user->host) 
             {
                 buffer[sidx++] = '@';
-                for(p = user->host; *p; p++)
+#ifdef USER_HOSTMASKING
+                if(IsUmodeH(from))
+                    p = user->mhost;
+                else
+#endif
+                    p = user->host;
+                for(; *p; p++)
                     buffer[sidx++] = *p;
                 flag = 1;
             }
@@ -588,10 +594,9 @@ void sendto_channel_butone(aClient *one, aClient *from, aChannel *chptr,
         if (acptr->from == one)
             continue; /* ...was the one I should skip */
 
-        if((confopts & FLAGS_SERVHUB) && IsULine(acptr) && (acptr->uplink->serv->uflags & ULF_NOCHANMSG))
+        if((confopts & FLAGS_SERVHUB) && IsULine(acptr) && (acptr->uplink->serv) && (acptr->uplink->serv->uflags & ULF_NOCHANMSG))
             continue; /* Don't send channel traffic to super servers */
 
-        i = acptr->from->fd;
         if (MyClient(acptr)) 
         {
             if(!didlocal)
@@ -604,7 +609,6 @@ void sendto_channel_butone(aClient *one, aClient *from, aChannel *chptr,
                     continue;
             
             send_message(acptr, sendbuf, didlocal, share_bufs[0]);
-            sentalong[i] = sent_serial;
         }
         else 
         {
@@ -622,6 +626,7 @@ void sendto_channel_butone(aClient *one, aClient *from, aChannel *chptr,
             if(check_fake_direction(from, acptr))
                     continue;
             
+            i = acptr->from->fd;
             if (sentalong[i] != sent_serial) 
             {
                 send_message(acptr, remotebuf, didremote, share_bufs[1]);
@@ -661,10 +666,12 @@ void sendto_channel_remote_butone(aClient *one, aClient *from, aChannel *chptr,
         if (acptr->from == one)
             continue; /* ...was the one I should skip */
 
-        if((confopts & FLAGS_SERVHUB) && IsULine(acptr) && (acptr->uplink->serv->uflags & ULF_NOCHANMSG))
+        if((confopts & FLAGS_SERVHUB) && IsULine(acptr) && (acptr->uplink->serv) && (acptr->uplink->serv->uflags & ULF_NOCHANMSG))
             continue; /* Don't send channel traffic to super servers */
 
-        i = acptr->from->fd;
+        if(acptr->fd == -2)
+            continue;
+
         if (!MyClient(acptr)) 
         {
             /*
@@ -681,6 +688,7 @@ void sendto_channel_remote_butone(aClient *one, aClient *from, aChannel *chptr,
             if(check_fake_direction(from, acptr))
                     continue;
             
+            i = acptr->from->fd;
             if (sentalong[i] != sent_serial) 
             {
                 send_message(acptr, remotebuf, didremote, share_buf);
@@ -938,7 +946,12 @@ void send_quit_to_common_channels(aClient *from, char *reason)
     INC_SERIAL
     
     msglen=sprintf(sendbuf,":%s!%s@%s QUIT :%s", from->name,
-                   from->user->username,from->user->host, reason);      
+                   from->user->username,
+#ifdef USER_HOSTMASKING
+                   IsUmodeH(from)?from->user->mhost:
+#endif
+                                                    from->user->host,
+                   reason);      
     sbuf_begin_share(sendbuf, msglen, &share_buf);
    
     if(from->fd >= 0)
@@ -984,7 +997,11 @@ void send_part_to_common_channels(aClient *from, char *reason)
         if (can_send(from, channels->value.chptr, reason)) 
         {
             msglen = sprintf(sendbuf,":%s!%s@%s PART %s",
-                             from->name,from->user->username,from->user->host,
+                             from->name,from->user->username,
+#ifdef USER_HOSTMASKING
+                             IsUmodeH(from)?from->user->mhost:
+#endif
+                                                              from->user->host,
                              channels->value.chptr->chname);
             sbuf_begin_share(sendbuf, msglen, &share_buf);
 
@@ -1032,13 +1049,13 @@ void sendto_channel_butlocal(aClient *one, aClient *from, aChannel *chptr,
         acptr = cm->cptr;
         if (acptr->from == one)
             continue;           /* ...was the one I should skip */
-        i = acptr->from->fd;
         if (!MyFludConnect(acptr)) 
         {
             /*
              * Now check whether a message has been sent to this remote
              * link already
              */
+            i = acptr->from->fd;
             if (sentalong[i] != sent_serial) 
             {
                 vsendto_prefix_one(acptr, from, pattern, vl);
@@ -1188,8 +1205,9 @@ void sendto_channel_butserv_me(aChannel *chptr, aClient *from, char *pattern, ..
 /*
  * sendto_channelopvoice_butserv_me
  * 
- * Send a message to all members of a channel that are connected to this
+ * Send a message to all opped or voiced members of a channel that are connected to this
  * server. Possibly hide the origin, if it's a server, with me.name if certain paranoia is on.
+ * IRC Operators will also receive this message (even if they're not opped/voiced).
  */
 void sendto_channelopvoice_butserv_me(aChannel *chptr, aClient *from, char *pattern, ...)
 {
@@ -1216,7 +1234,7 @@ void sendto_channelopvoice_butserv_me(aChannel *chptr, aClient *from, char *patt
     {
         if (MyConnect(acptr = cm->cptr))
         {
-            if(!is_chan_opvoice(acptr, chptr)) continue;
+            if(!is_chan_opvoice(acptr, chptr) && !IsAnOper(acptr)) continue;
             if (!didlocal)
             {
                 didlocal = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
@@ -1631,10 +1649,16 @@ void sendto_prefix_one(aClient *to, aClient *from, char *pattern, ...)
                     for(idx = user->username; *idx; idx++)
                         sender[sidx++] = *idx;
                 }
-                if (*user->host && !MyConnect(from)) 
+                if (*user->host) 
                 {
                     sender[sidx++] = '@';
-                    for(idx = user->host; *idx; idx++)
+#ifdef USER_HOSTMASKING
+                    if(IsUmodeH(from))
+                        idx = user->mhost;
+                    else
+#endif
+                        idx = user->host;
+                    for(; *idx; idx++)
                         sender[sidx++] = *idx;
                     flag = 1;
                 }
@@ -1740,10 +1764,16 @@ void vsendto_prefix_one(aClient *to, aClient *from, char *pattern, va_list vl)
                     for(idx = user->username; *idx; idx++)
                         sender[sidx++] = *idx;
                 }
-                if (*user->host && !MyConnect(from)) 
+                if (*user->host) 
                 {
                     sender[sidx++] = '@';
-                    for(idx = user->host; *idx; idx++)
+#ifdef USER_HOSTMASKING
+                    if(IsUmodeH(from))
+                        idx = user->mhost;
+                    else
+#endif
+                        idx = user->host;
+                    for(; *idx; idx++)
                         sender[sidx++] = *idx;
                     flag = 1;
                 }
@@ -2085,7 +2115,7 @@ void sendto_channelflags_butone(aClient *one, aClient *from, aChannel *chptr,
         if (acptr->from == one || !(cm->flags & flags))
             continue;
 
-        if((confopts & FLAGS_SERVHUB) && IsULine(acptr) && (acptr->uplink->serv->uflags & ULF_NOCHANMSG))
+        if((confopts & FLAGS_SERVHUB) && IsULine(acptr) && (acptr->uplink->serv) && (acptr->uplink->serv->uflags & ULF_NOCHANMSG))
             continue; /* Don't send channel traffic to super servers */
 
         if (MyConnect(acptr))
